module Nicolai where

open import Level

id : {i : Level}{A : Set i} → A → A
id = λ x → x

infix 1 _≡_

data _≡_ {i : Level}{A : Set i}(a : A) : A → Set where
  refl : a ≡ a

J : {i : Level} {A : Set i}
    {j : Level} (P : (a b : A) → a ≡ b → Set j) (h : (a : A) → P a a refl)
    {a b : A} (q : a ≡ b) → P a b q
J P h {a} refl = h a

subst : {i : Level} {A : Set i} {a b : A} (q : a ≡ b)
        {j : Level} (P : A → Set j) → P a → P b
subst q P = J (λ a b _ → P a → P b) (λ _ x → x) q

trans : {i : Level} {A : Set i} {a b c : A} → a ≡ b → b ≡ c → a ≡ c
trans {a = a} p q = subst q (λ x → a ≡ x) p

resp : {i j : Level} {A : Set i} {B : Set j} (f : A → B) →
       {a b : A} → a ≡ b → f a ≡ f b
resp f {a} p = subst p (λ x → f a ≡ f x) refl

postulate
  nicolai :
    {i : Level} {A : Set i}
    {j : Level} (P : (a b : A) → a ≡ b → Set j) (h : (a : A) → P a a refl)
    {a b : A} (q : a ≡ b) →
    (p : P a b q ≡ P a a refl) → subst p id (J P h q) ≡ h a

postulate
  ext : {A : Set}{B : A → Set} {f g : (x : A) → B x} →
        ((x : A) → f x ≡ g x) → f ≡ g

DEq : {A : Set} {a b : A} (p : a ≡ b) (P : A → Set) (x : P a) (y : P b) → Set
DEq p P x y = subst p P x ≡ y

data Bool : Set where
  true false : Bool

boolApp : {P : Bool → Set} → ((x : Bool) -> P x) → (x : Bool) → P x
boolApp f = λ { true → f true; false → f false }

not : Bool → Bool
not true = false
not false = true

infix 5 _∘_

_∘_ : {A B C : Set} → (B → C) → (A → B) → A → C
f ∘ g = λ x → f (g x)

notNotIdx : (x : Bool) → x ≡ not (not x)
notNotIdx true  = refl
notNotIdx false = refl

notNotId : id ≡ not ∘ not
notNotId = ext notNotIdx

T : Bool → Set
T true = Bool
T false = Bool → Bool

P : (Bool → Bool) → Set
P f = T (f false)

-- consequences of Nicolai's reduction rule
postulate
  -- P id = P (not ∘ not), hence
  nx     : (x : Bool) → subst notNotId P id x ≡ x
  ntrue  : subst notNotId P id true ≡ true
  nfalse : subst notNotId P id false ≡ false

-- subst notNotId P id ≡ not ∘ not
nni : DEq notNotId P id (not ∘ not)
-- nni = ext (boolApp nx)
-- nni = ext (λ { true -> nx true; false -> nx false })
nni = ext (λ x → trans (nx x) (notNotIdx x))
-- with Nicolai's rule, this proof is the same as notNotId

data Relate : (Bool → Bool) → (Bool → Bool) → Set where
  relate : Relate id (not ∘ not)

lemma : {f g : Bool → Bool} (r : Relate f g) → f ≡ g
lemma relate = notNotId

{-
lem1  : {f g : Bool → Bool} (r : Relate f g) → DEq (lemma r) P f g
lem1 relate = nni
-}


